#if defined(VMPI)

#include <globdev.h>

#undef DEBUG_FN_NAME
#define DEBUG_FN_NAME MPID_Comm_init
int MPID_Comm_init(
    struct MPIR_COMMUNICATOR *		oldcomm,
    struct MPIR_COMMUNICATOR *		newcomm)
{
    int					rc;
    int					i;
    int					vlnp;

    DEBUG_FN_ENTRY(DEBUG_MODULE_COMM);

    rc = MPI_SUCCESS;

    if (newcomm != NULL)
    {
	newcomm->vmpi_comm = NULL;
	newcomm->lrank_to_vlrank = NULL;
	newcomm->vlrank_to_lrank = NULL;
	newcomm->vgrank_to_vlrank = NULL;
    }

    /*
     * If we can't use the vendor MPI to communicate then we are done
     */
    if (VMPI_MyWorldRank < 0
	|| (oldcomm != NULL && oldcomm->vmpi_comm == NULL))
    {
	goto fn_exit;
    }

    /*
     * If we aren't part of the new communicator, then we don't need to
     * initialze any of the data structures.  However, we still need to
     * participate in the mp_comm_split().
     */
    if (newcomm == NULL)
    {
	vlnp = 0;
	goto skip_newcomm_init;
    }
    
    /*
     * Establish a mapping from global ranks in the vendor MPI_COMM_WORLD to
     * local ranks in the new MPICH communicator, and a mapping from local
     * ranks in the new MPICH communicator to local ranks in the vendor
     * communicator
     */
    newcomm->vgrank_to_vlrank = (int *)
	globus_libc_malloc(VMPI_MyWorldSize * sizeof(int));
    if (newcomm->vgrank_to_vlrank == NULL)
    {
	rc = MPI_ERR_NOMEM;
	goto fn_abort;
    }
    for (i = 0; i < VMPI_MyWorldSize; i++)
    {
	newcomm->vgrank_to_vlrank[i] = -1;
    }

    newcomm->lrank_to_vlrank = (int *)
	globus_libc_malloc(newcomm->np * sizeof(int));
    if (newcomm->lrank_to_vlrank == NULL)
    {
	rc = MPI_ERR_NOMEM;
	goto fn_abort;
    }

    vlnp = 0;
    for (i = 0; i < newcomm->np; i++)
    {
	int				vgrank;

	vgrank = VMPI_GRank_to_VGRank[newcomm->lrank_to_grank[i]];
	if (vgrank >= 0)
	{
	    assert(vgrank < VMPI_MyWorldSize);
	    newcomm->vgrank_to_vlrank[vgrank] = vlnp;
	    newcomm->lrank_to_vlrank[i] = vlnp;
	    DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_MISC,
			 ("newcomm->lrank_to_vlrank[%d]=%d\n",
			  i, vlnp));

	    vlnp++;
	}
	else
	{
	    DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_MISC,
			 ("newcomm->vgrank_to_vlrank[%d]=-1\n", vlnp));
	    newcomm->lrank_to_vlrank[i] = -1;
	}
    }

#   if DEBUG_CHECK(DEBUG_MODULE_COMM, DEBUG_INFO_MISC)
    {
	for (i = 0; i < VMPI_MyWorldSize; i++)
	{
	    DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_MISC,
			 ("newcomm->vgrank_to_vlrank[%d]=%d\n",
			  i, newcomm->vgrank_to_vlrank[i]));

	}
    }
#   endif
    
    DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_MISC,
		 ("newcomm->np=%d newcomm->vlnp=%d\n", newcomm->np, vlnp));

    if (vlnp == 0)
    {
	/*
	 * We are not part of the new communicator, so we should free up our
	 * data structures
	 */
	g_free(newcomm->vgrank_to_vlrank);
	g_free(newcomm->lrank_to_vlrank);

	newcomm->vgrank_to_vlrank = NULL;
	newcomm->lrank_to_vlrank = NULL;

	goto skip_newcomm_init;
    }

    /*
     * Establish a mapping from local ranks in the vendor communicator to
     * local ranks in the new MPICH communicator
     */
    newcomm->vlrank_to_lrank = (int *)
	globus_libc_malloc(vlnp * sizeof(int));
    if (newcomm->vlrank_to_lrank == NULL)
    {
	rc = MPI_ERR_NOMEM;
	goto fn_abort;
    }

    for (i = 0; i < newcomm->np; i++)
    {
	int				vlrank;
	int				vgrank;

	vlrank = newcomm->lrank_to_vlrank[i];
	if (vlrank >= 0)
	{
	    assert(vlrank < vlnp);
	    newcomm->vlrank_to_lrank[vlrank] = i;
	}
    }
    
#   if DEBUG_CHECK(DEBUG_MODULE_COMM, DEBUG_INFO_MISC)
    {
	for (i = 0; i < vlnp; i++)
	{
	    DEBUG_PRINTF_NOCHECK(
		("newcomm->vlrank_to_lrank[%d]=%d\n",
		 i, newcomm->vlrank_to_lrank[i]));
	}
    }
#   endif
    
    newcomm->vmpi_comm = (int *)
	globus_libc_malloc(mp_comm_get_size());
    if (newcomm->vmpi_comm == NULL)
    {
	rc = MPI_ERR_NOMEM;
	goto fn_abort;
    }

  skip_newcomm_init:
    
    if ((oldcomm == NULL || oldcomm->comm_type == MPIR_INTRA)
	&& (newcomm == NULL || newcomm->comm_type == MPIR_INTRA))
    {
	int				color;
	int				key;
	
	if (vlnp > 0)
	{
	    color = newcomm->lrank_to_grank[newcomm->vlrank_to_lrank[0]];
	    key = newcomm->lrank_to_vlrank[newcomm->local_rank];
	}
	else
	{
	    color = VMPI_UNDEFINED;
	    key = 0;
	}
	
	rc = vmpi_error_to_mpich_error(
	    mp_comm_split(oldcomm ? oldcomm->vmpi_comm : NULL,
			  color,
			  key,
			  newcomm ? newcomm->vmpi_comm : NULL));
	if (rc != MPI_SUCCESS)
	{
	    goto fn_abort;
	}
    }
    else if (oldcomm->comm_type == MPIR_INTRA
	     && newcomm->comm_type == MPIR_INTER)
    {
	/* XXX: must be performing a MPI_INTERCOMM_CREATE() */
	DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_FAILURE,
		     ("VMPI inter-communicators have not be implemented\n"));
	abort();
    }
    else if (oldcomm->comm_type == MPIR_INTER
	     && newcomm->comm_type == MPIR_INTRA)
    {
	/* XXX: must be performing a MPI_INTERCOMM_MERGE() */
	DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_FAILURE,
		     ("VMPI inter-communicators have not be implemented\n"));
	abort();
    }
    else if (oldcomm->comm_type == MPIR_INTER
	     && newcomm->comm_type == MPIR_INTER)
    {
	/* XXX: must be duplicating an existing inter-communicator */
	DEBUG_PRINTF(DEBUG_MODULE_COMM, DEBUG_INFO_FAILURE,
		     ("VMPI inter-communicators have not be implemented\n"));
	abort();
    }
    else
    {
    }

    goto fn_exit;
    
  fn_abort:
    g_free(newcomm->vmpi_comm);
    g_free(newcomm->vlrank_to_lrank);
    g_free(newcomm->lrank_to_vlrank);
    g_free(newcomm->vgrank_to_vlrank);
    
    newcomm->vmpi_comm = NULL;
    newcomm->lrank_to_vlrank = NULL;
    newcomm->vlrank_to_lrank = NULL;
    newcomm->vgrank_to_vlrank = NULL;

  fn_exit:
    DEBUG_FN_EXIT(DEBUG_MODULE_COMM);
    return rc;
}

#undef DEBUG_FN_NAME
#define DEBUG_FN_NAME MPID_Comm_free
int MPID_Comm_free(
    struct MPIR_COMMUNICATOR *		comm)
{
    DEBUG_FN_ENTRY(DEBUG_MODULE_COMM);

    if (comm->vmpi_comm != NULL)
    {
	mp_comm_free(comm->vmpi_comm);
    }

    g_free(comm->vmpi_comm);
    g_free(comm->vlrank_to_lrank);
    g_free(comm->lrank_to_vlrank);
    g_free(comm->vgrank_to_vlrank);
    
    comm->vmpi_comm = NULL;
    comm->lrank_to_vlrank = NULL;
    comm->vlrank_to_lrank = NULL;
    comm->vgrank_to_vlrank = NULL;

  fn_exit:
    DEBUG_FN_EXIT(DEBUG_MODULE_COMM);
    return MPI_SUCCESS;
}

#endif /* defined(VMPI) */
